#  Copyright (c) 2019 JD Williams
#
#  This file is part of Firefly, a Python SOA framework built by JD Williams. Firefly is free software; you can
#  redistribute it and/or modify it under the terms of the GNU General Public License as published by the
#  Free Software Foundation; either version 3 of the License, or (at your option) any later version.
#
#  Firefly is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
#  implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
#  Public License for more details. You should have received a copy of the GNU Lesser General Public
#  License along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#  You should have received a copy of the GNU General Public License along with Firefly. If not, see
#  <http://www.gnu.org/licenses/>.

from __future__ import annotations

import importlib
from abc import ABC
import firefly_di as di
import firefly.domain as ffd


class ModuleBase(ABC, ffd.LoggerAware):
    def load_container(self, name: str, bus: ffd.MessageBus):
        try:
            self.debug('Attempting to import module {}.application'.format(name))
            module = importlib.import_module('{}.application'.format(name))
            container_class = getattr(module, 'Container')
            self.debug('Container imported successfully')
        except (ModuleNotFoundError, AttributeError):
            self.debug('Failed to load application module for {}. Ignoring.', name)

            class EmptyContainer(di.Container):
                pass
            container_class = EmptyContainer

        self.debug('Injecting message bus into container')
        container_class.message_bus = bus
        container_class.__annotations__['message_bus'] = ffd.MessageBus
        container = container_class()

        return container
