#  Copyright (c) 2019 JD Williams
#
#  This file is part of Firefly, a Python SOA framework built by JD Williams. Firefly is free software; you can
#  redistribute it and/or modify it under the terms of the GNU General Public License as published by the
#  Free Software Foundation; either version 3 of the License, or (at your option) any later version.
#
#  Firefly is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
#  implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
#  Public License for more details. You should have received a copy of the GNU Lesser General Public
#  License along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#  You should have received a copy of the GNU General Public License along with Firefly. If not, see
#  <http://www.gnu.org/licenses/>.
#
#  This file is part of Firefly, a Python SOA framework built by JD Williams. Firefly is free software; you can
#  redistribute it and/or modify it under the terms of the GNU General Public License as published by the
#  Free Software Foundation; either version 3 of the License, or (at your option) any later version.
#
#  Firefly is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
#  implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
#  Public License for more details. You should have received a copy of the GNU Lesser General Public
#  License along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#  You should have received a copy of the GNU General Public License along with Firefly. If not, see
#  <http://www.gnu.org/licenses/>.

from __future__ import annotations

from multiprocessing import Pipe
from multiprocessing.context import Process
from typing import Callable, Iterable

from ..core.domain_service import DomainService


def wrapper(cb):
    def _wrapper(*args):
        conn = args[-1]
        try:
            conn.send(cb(*args[0:-1]))
            conn.close()
        except Exception as e:
            conn.send(e)
        finally:
            conn.close()
    return _wrapper


class BatchProcess(DomainService):
    def __call__(self, cb: Callable, args: Iterable, timeout: int = None):
        processes = []
        connections = []

        for arg_set in args:
            parent, child = Pipe(duplex=False)
            connections.append(parent)
            process = Process(target=wrapper(cb), args=arg_set + (child,))
            processes.append(process)

        for p in processes:
            p.start()

        ret = []
        for c in connections:
            ret.append(c.recv())

        for p in processes:
            p.join(timeout=timeout)

        return ret
